# __name__ = "chemml"
__version__ = "1.0"
__author__ = ["Aditya Sonpal (adityaso@buffalo.edu)", "Garuav Vishwakarma (gvishwak@buffalo.edu) ", "Aatish Pradhan (aatishpr@buffalo.edu)","Mojtaba Haghighatlari (mojtabah@buffalo.edu)", "Johannes Hachmann (hachmann@buffalo.edu)"]


# import sys
# sys.dont_write_bytecode = True
