"""
Collection of simple preprocessing wrappers only for the sake of user friendliness.
"""

from .miscellaneous import SplitColumns
from .miscellaneous import SaveCSV
from .miscellaneous import SaveFile
from .miscellaneous import SaveHDF5